/*
 * TUX - Integrated HTTP layer and Object Cache
 *
 * Copyright (C) 2000, Ingo Molnar <mingo@redhat.com>
 *
 * CAD.c: Implementation of the SPECweb99 dynamic application via
 *        the HTTP trusted-API.
 */

#include <net/http.h>

static inline int send_reply_head (http_req_t *req, size_t body_size);
static inline int send_reply_tail (http_req_t *req);
static int send_err (http_req_t *req, char *message);

/*
 * We memory-map the head of the postlog file so that we
 * do not have to call write() every time we update it.
 * The postlog record is written out to disk on every POST
 * request, and the record counter (in the memory mapped
 * buffer) is updated atomically as well.
 */
#define POSTLOG "/tmp/postlog"

static HTTP_DECLARE_MUTEX(postlog_sem);
static struct http_file *post_file;
static int postlog_count;
static char *postlog_head;
static struct http_page *postlog_head_page;

#define CAD_TAG_HEAD	"<!WEB99CAD><IMG SRC=\"/file_set/"
#define CAD_TAG_BODY	"dirNNNNN/classX_Y"
#define CAD_TAG_TAIL	"\"><!/WEB99CAD>"

#define CAD_TAG CAD_TAG_HEAD CAD_TAG_BODY CAD_TAG_TAIL

#define CAD_TAG_BODY_POS (sizeof(CAD_TAG_HEAD)-1)
#define CAD_TAG_BODY_LEN (sizeof(CAD_TAG_BODY)-1)
#define CAD_TAG_TAIL_LEN (sizeof(CAD_TAG_TAIL)-1)
#define CAD_TAG_LEN (sizeof(CAD_TAG)-1)

typedef struct CAD_struct {
	int user_id;
	int last_ad;
	char ad_filename [100];
	int reply_cookies_len;
	char reply_cookies[MAX_COOKIE_LEN];
} CAD_t;

static inline int is_class12 (char *str)
{
	unsigned char *tmp;

	tmp = strstr(str, "class");
	if (tmp) {
		tmp += sizeof("class")-1;
		if ((tmp[0] != '1') && (tmp[0] != '2'))
			return 0;
		return 1;
	}
	return 0;
}

/*
 * This function reads the object, scans the temporary buffer for
 * the SPECweb99 tag string, does CAD replacement and sends the
 * result out to the client.
 */
static inline int scan_send_file (http_req_t *req, CAD_t *CADp)
{
	char *tmpbuf, *target;
	int size, left, ret;
	mm_segment_t oldmm;

	if (!CADp || !req->query || !is_class12(req->query))
		return http_send_object(req, 0, 0);

	size = req->urlo->filelen;
	tmpbuf = req->ti->output_buffer;
	oldmm = get_fs(); set_fs(KERNEL_DS);
	ret = http_read(req->urlo, tmpbuf);
	set_fs(oldmm);

	if (ret != size)
		return send_err(req, "CAD: error while reading object file!\n");
	tmpbuf[size] = 0; // zero-delimited string

	target = tmpbuf;
	left = size - CAD_TAG_LEN + 1;

	for (;;) {
		target = memchr(target, CAD_TAG[0], left);
		if (!target)
			break; // no such character left

		if (memcmp(target, CAD_TAG, CAD_TAG_LEN)) {
			target++; // skip past the '<'
			left--;
			continue;
		}
		target += CAD_TAG_BODY_POS;
		memcpy(target, CADp->ad_filename, CAD_TAG_BODY_LEN);
		target += CAD_TAG_BODY_LEN + CAD_TAG_TAIL_LEN;
		left -= CAD_TAG_LEN;
	}

	http_send_client(req, tmpbuf, size, 0);

	return size;
}


typedef struct user_dem_s {
	unsigned int dem;
} user_dem_t;

static int max_userid;
static user_dem_t *user_dem = NULL;
static struct http_direntry *user_pers_dentry;

#define USER_PERS_FILE "User.Personality"
#define USER_PERS_RECLEN 15

typedef struct ad_s {
	unsigned int dem;
	unsigned int gender_weight;
	unsigned int age_weight;
	unsigned int region_weight;
	unsigned int interest_1_weight;
	unsigned int interest_2_weight;
	unsigned int min_match;
	unsigned int expires;
} ad_t;

static int max_adid;
static ad_t *ad = NULL;

#define AD_FILE "Custom.Ads"
#define AD_RECLEN 39

static int read_custom_ads (http_req_t *req)
{
	struct http_file *file;
	int ret, len, err = -2;
	char *buf = NULL, *tmp;
	struct http_direntry *dentry;
	unsigned int adid, i, dem, min_match, weight, expires;


	dentry = http_lookup_direntry(AD_FILE, &docroot, 0);
	if (http_direntry_error(dentry))
		goto error;
	file = http_direntry_open(dentry, O_RDONLY, 0);
	if (!file)
		goto error;
	len = http_file_size(file);
	if (!len)
		goto error;
	if (ad) {
		http_free(ad, ALLOC_AD_FILE);
		ad = NULL;
	}
	max_adid = len/AD_RECLEN + 1;
	ad = http_malloc(max_adid * sizeof(ad_t), ALLOC_AD_FILE);
	buf = http_malloc(len, ALLOC_ADTMPBUF);
	if (!ad || !buf)
		goto error;

	ret = http_read_file(file, buf, len);
	http_close_file(file);
	if (ret != len)
		goto error;

/*
 * Sample ad record:
	"   54 24808100    97F61  75  952393980\n"
 */

	tmp = buf;
	i = 0;
	for (tmp = buf; tmp != buf+len; tmp++) {

		while (*tmp == ' ') tmp++;
		adid = simple_strtoul(tmp, &tmp, 10);
		if (adid != i)
			goto error;
		if (adid >= max_adid)
			goto error;
		i++;
		if (*tmp != ' ')
			goto error;
		tmp++;
		while (*tmp == ' ') tmp++;
		dem = simple_strtoul(tmp, &tmp, 16);
		tmp++;
		while (*tmp == ' ') tmp++;
		weight = simple_strtoul(tmp, &tmp, 16);
		while (*tmp == ' ') tmp++;
		min_match = simple_strtoul(tmp, &tmp, 10);
		while (*tmp == ' ') tmp++;
		expires = simple_strtoul(tmp, &tmp, 10);
		if (*tmp != '\n')
			goto error;
		ad[adid].dem = dem;

		ad[adid].gender_weight		= (weight & 0x000f0000) >> 16;
		ad[adid].age_weight		= (weight & 0x0000f000) >> 12;
		ad[adid].region_weight		= (weight & 0x00000f00) >> 8;
		ad[adid].interest_1_weight	= (weight & 0x000000f0) >> 4;
		ad[adid].interest_2_weight	= (weight & 0x0000000f);

		ad[adid].min_match = min_match;
		ad[adid].expires = expires;

	}
	err = 0;
error:
	if (buf)
		http_free(buf, ALLOC_ADTMPBUF);
	if (err)
		return send_err(req, "CAD: error while reading & parsing the ad file.\n");
	return err;
}

static int read_user_personality (http_req_t *req)
{
	struct http_file *file;
	int ret, len, err = -2;
	char *buf = NULL, *tmp;
	unsigned int uid, i, dem;
	struct http_direntry *dentry;

	dentry = http_lookup_direntry(USER_PERS_FILE, &docroot, 0);
	if (http_direntry_error(dentry))
		goto error;
	file = http_direntry_open(dentry, O_RDONLY, 0);
	if (!file)
		goto error;
	len = http_file_size(file);
	if (!len)
		goto error;
	if (user_dem) {
		http_free(user_dem, ALLOC_USERDEM);
		user_dem = NULL;
	}
	max_userid = len/USER_PERS_RECLEN + 1;
	user_dem = http_malloc(max_userid * sizeof(user_dem_t), ALLOC_USERDEM);
	buf = http_malloc(len, ALLOC_USERDEM_TMPBUF);
	if (!user_dem || !buf) {
		goto error;
	}

	ret = http_read_file(file, buf, len);
	http_close_file(file);
	if (ret != len) {
		goto error;
	}

	i = 0;
	for (tmp = buf; tmp != buf+len; tmp++) {
		if (*tmp == ' ')
			continue;
		uid = simple_strtoul(tmp, &tmp, 10);
		if (uid != i)
			goto error;
		if (uid >= max_userid)
			goto error;
		i++;
		if (*tmp != ' ')
			goto error;
		while (*tmp == ' ') tmp++;
		dem = simple_strtoul(tmp, &tmp, 16);
		if (*tmp != '\n')
			goto error;
		user_dem[uid].dem = dem;
	}
	err = 0;
error:
	if (buf)
		http_free(buf, ALLOC_USERDEM_TMPBUF);
	if (err)
		return send_err(req, "CAD: error while reading & parsing the user file.\n");
	return err;
}

#define MAX_CUSTOM_ADS 360

static inline int find_ad (int user_id, int last_ad, int *weight_p)
{
	int adid, weight = 0, dem;

	for (adid = last_ad + 1; adid != last_ad; adid++) {
		if (adid >= MAX_CUSTOM_ADS)
			adid = 0;

		dem = user_dem[user_id].dem & ad[adid].dem;
		weight = 0;

		if (dem & 0x30000000)
			weight += ad[adid].gender_weight;
		if (dem & 0x0f000000)
			weight += ad[adid].age_weight;
		if (dem & 0x00f00000)
			weight += ad[adid].region_weight;
		if (dem & 0x000ffc00)
			weight += ad[adid].interest_1_weight;
		if (dem & 0x000003ff)
			weight += ad[adid].interest_2_weight;
		if (weight >= ad[adid].min_match)
			break;
	}

	*weight_p = weight;
	return adid;
}

static unsigned int last_mtime = 0;

static int reread_files (http_req_t *req)
{
	int ret = -2;
	struct http_direntry *dentry;

	http_dput(user_pers_dentry);
	dentry = http_lookup_direntry(USER_PERS_FILE, &docroot, 0);
	if (http_direntry_error(dentry))
		goto error;
	user_pers_dentry = dentry;

	if (http_mtime(dentry) != last_mtime) {
		void *tmp = user_dem;
		user_dem = NULL;
		http_free(tmp, ALLOC_USERDEM);
		if (read_user_personality(req))
			goto error;
		if (read_custom_ads(req))
			goto error;
		last_mtime = http_mtime(dentry);
	}
	ret = 0;

error:
	return ret;
}

static inline int custom_ad_rotate (http_req_t *req, CAD_t *CADp)
{
	int adid, weight, expired, err;
	int user_id, last_ad;
	time_t now;

	user_id = CADp->user_id;
	last_ad = CADp->last_ad;

	if (http_direntry_error(user_pers_dentry) ||
			(http_mtime(user_pers_dentry) != last_mtime)) {
		err = reread_files(req);
		if (err)
			return err;
	}

	/*
	 * Any error in either reading or parsing of the files results
	 * in a returned -1 adid.
	 */
	adid = -1;
	expired = 1;
	weight = 0;

	adid = find_ad(user_id, last_ad, &weight);
	if (adid < 0)
		goto error;
	now = http_time();
	if (now <= ad[adid].expires)
		expired = 0;

error:
	CADp->reply_cookies_len = sprintf(CADp->reply_cookies,
		"found_cookie=Ad_id=%d&Ad_weight=%d&Expired=%d",
			adid, weight, expired);

	sprintf(CADp->ad_filename, "dir%05d/class%d_%d",
		adid / 36, ((adid % 36) / 9), adid % 9);
	return 0;
}


#define TOKEN_EQUAL(input,token) \
		(!memcmp(input, token, sizeof(token)-1))

#define PARSE_STRING(token,input,output)				\
	({								\
		int __ret = 0;						\
		if (TOKEN_EQUAL(input, token)) {			\
			char *tmp;					\
									\
			input += sizeof(token)-1;			\
			tmp = output;					\
			while (*input && *input != '&' &&		\
						*input != ',')		\
				*tmp++ = *input++;			\
			*tmp = 0;					\
			__ret = 1;					\
		}							\
		__ret;							\
	})

#define PARSE_UINT(token,input,output)					\
	({								\
		int __ret = 0;						\
		if (TOKEN_EQUAL(input, token)) {			\
									\
			input += sizeof(token)-1;			\
			output = simple_strtoul(input, &input, 10);	\
			__ret = 1;					\
		}							\
		__ret;							\
	})

static int init_postlog_file (http_req_t *req)
{
	char buf[400], *tmp;
	int ret;

	if (post_file)
		http_close_file(post_file);
	post_file = http_open_file(POSTLOG, O_CREAT|O_TRUNC|O_APPEND|O_RDWR);
	if (!post_file)
		return send_err(req, "CAD: could not open POST-log!\n");
	postlog_count = 0;
	tmp = buf;
	tmp += sprintf(tmp, "%10d\n", 0);
	ret = http_write_file(post_file, buf, tmp-buf);
	if (ret != tmp-buf)
		return send_err(req, "CAD: POST-log write error!\n");
	postlog_head_page = http_mmap_page(post_file, postlog_head, 0);
	if (!postlog_head_page)
		return send_err(req, "CAD: POST-log mmap error!\n");
	return 0;
}

#define COMMAND_STRING "command/"
#define COMMAND_RESET "Reset"
#define COMMAND_FETCH "Fetch"

static int do_reset (http_req_t *req, char *query)
{
	char maxload [20], pttime[20], maxthread[20],
			exp1[20], exp2[20], urlroot [100];
	char tmpstr1[256], tmpstr2[256];

	http_sleep(1);
	if (!PARSE_STRING("&maxload=", query, maxload))
		return send_err(req,"CAD: invalid &maxload field!\n");
	if (!PARSE_STRING("&pttime=", query, pttime))
		return send_err(req,"CAD: invalid &pttime field!\n");
	if (!PARSE_STRING("&maxthread=", query, maxthread))
		return send_err(req,"CAD: invalid &maxthread field!\n");
	if (!PARSE_STRING("&exp=", query, exp1))
		return send_err(req,"CAD: invalid &exp1 field!\n");
	if (!PARSE_STRING(",", query, exp2))
		return send_err(req,"CAD: invalid &exp2 field!\n");
	if (!PARSE_STRING("&urlroot=", query, urlroot))
		return send_err(req,"CAD: invalid &urlroot field!\n");


	strcpy(tmpstr1, http_docroot); strcat(tmpstr1, "/upfgen99");
	strcpy(tmpstr2, http_docroot); strcat(tmpstr2, "/cadgen99");
#define TOPDIR http_docroot
#define UPFGEN tmpstr1
#define CADGEN tmpstr2

	{
		char *argv_upfgen[] = { UPFGEN, "-C", TOPDIR, "-n", maxload,
					"-t", maxthread, NULL};
		char *argv_cadgen[] = { CADGEN, "-C", TOPDIR, "-e", pttime,
					"-t", maxthread, exp1, exp2, NULL};
		char * envp[] = { "HOME=/", "TERM=linux",
			"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL };

		if (http_exec_process(UPFGEN, argv_upfgen, envp, 0, 0, 1) < 0)
			return send_err(req,"CAD: could not execute UPFGEN!\n");

		if (http_exec_process(CADGEN, argv_cadgen, envp, 0, 0, 1) < 0)
			return send_err(req,"CAD: could not execute CADGEN!\n");
	}
	/*
	 * Clear post log
	 */
	http_down(&postlog_sem);
	init_postlog_file(req);
	http_up(&postlog_sem);

	/*
	 * mtime has a 1 second resolution, sleep 1 second so that
	 * the check for modified User.Personality and Custom.Ads
	 * files notices multiple resets correctly.
	 */
	http_sleep(1);

	req->bytes_sent = send_reply_head(req, 0);
	req->bytes_sent += send_reply_tail(req);

	return -2;
}

#define BLOCKLEN 512

static int send_postlog (http_req_t *req)
{
	char buf [BLOCKLEN];
	struct http_file *file;
	int len, total, bytes;

	file = http_open_file(POSTLOG, O_RDONLY);
	if (http_file_error(file))
		return send_err(req, "CAD: no POST-log file!\n");

	req->body_len = http_file_size(file);
	bytes = send_reply_head(req, req->body_len);
	http_down(&postlog_sem);
	total = 0;
	do {
		len = http_read_file(file, buf, BLOCKLEN);
		if (len <= 0)
			break;
		bytes += http_send_client(req, buf, len, 0);
		total += len;
	} while (len == BLOCKLEN);

	http_close_file(file);
	http_up(&postlog_sem);
	bytes += send_reply_tail(req);
	req->bytes_sent = bytes;
	req->keep_alive = 0;

	return -2;
}

static int do_command (http_req_t *req, char *query)
{
	if (TOKEN_EQUAL(query, COMMAND_RESET))
		return do_reset(req, query + sizeof(COMMAND_RESET)-1);
	if (TOKEN_EQUAL(query, COMMAND_FETCH))
		return send_postlog(req);
	return send_err(req,"CAD: got invalid command!\n");
}

static CAD_t * parse_GET_cookies (http_req_t *req)
{
	int uid, last_ad;
	CAD_t *CADp;
	char *tmp;

	if (!req->cookies_len) {
		return NULL;
	}

	CADp = http_malloc(sizeof(CAD_t), ALLOC_REQ_PRIVATE);
	CADp->reply_cookies_len = 0;
	req->private = (void *) CADp;

	tmp = req->cookies + sizeof("my_cookie=user_id=")-1;
	uid = simple_strtoul(tmp, &tmp, 10) - 10000;

	tmp += sizeof("&last_ad=")-1;
	last_ad = simple_strtoul(tmp, &tmp, 10);

	CADp->user_id = uid;
	CADp->last_ad = last_ad;

	return CADp;
}

static int do_POST (http_req_t *req)
{
	int dir = -1, class = -1, num = -1, client = -1;
	char buf[400], *tmp;
	char urlroot[100];
	CAD_t *CADp;
	char *curr;
	int ret;

	CADp = parse_GET_cookies(req);

	curr = req->post_data;
	if (!curr)
		goto parse_error;

#define POST_URLROOT "urlroot="
#define POST_CLASS "class="
#define POST_CLIENT "client="
#define POST_DIR "dir="
#define POST_NUM "num="

	for (;;) {
		switch (*curr) {
			case 'u':
				if (PARSE_STRING( POST_URLROOT, curr, urlroot))
					continue;
				goto parse_error;
			case 'c':
				if (PARSE_UINT( POST_CLASS, curr, class))
					continue;
				if (PARSE_UINT( POST_CLIENT, curr, client))
					continue;
				goto parse_error;
			case 'd':
				if (PARSE_UINT( POST_DIR, curr, dir))
					continue;
				goto parse_error;
			case 'n':
				if (PARSE_UINT( POST_NUM, curr, num))
					continue;
				goto parse_error;
			case '&':
				curr++;
				continue;
			case 0:
				goto out;
			default:
				goto parse_error;
		}
		goto parse_error;
	}
out:
	if (!CADp)
		goto parse_error;
	tmp = CADp->ad_filename;
	tmp += sprintf(tmp, "%sdir%05d/class%d_%d", urlroot, dir, class, num);

	/*
	 * Aquire semaphore guaranteeing atomic operations
	 * on the postlog file.
	 */
	http_down(&postlog_sem);
	if (!post_file)
		if (init_postlog_file(req))
			return 0;

	postlog_count++;
	tmp = postlog_head;
	tmp += sprintf(tmp, "%10d", postlog_count);
	*tmp = '\n';

	tmp = buf;
	tmp += sprintf(tmp, "%10d %10ld %10d %5d %2d %2d %10d %-60.60s %10d %10d\n", postlog_count, http_time(), http_getpid(), dir, class, num, client, CADp->ad_filename, http_getpid(), CADp->user_id + 10000);

	ret = http_write_file(post_file, buf, tmp-buf);
	http_up(&postlog_sem);

	if (ret != tmp-buf)
		goto write_error;

	CADp->reply_cookies_len = sprintf(CADp->reply_cookies,
		"my_cookie=%u", 10000 + CADp->user_id);
	return 0;

parse_error:
	return send_err(req,"CAD: error while parsing POST request!\n");

write_error:
	return send_err(req, "CAD: POST-log write error!\n");
}

static int query_CAD (http_req_t *req)
{
	urlo_t *urlo = NULL;
	int ret = 0;
	CAD_t *CADp;
	int missed;


	if (req->method == METHOD_POST) {
		ret = do_POST(req);
		if (ret)
			return ret;
		CADp = (CAD_t *)req->private;
		req->objectname = CADp->ad_filename;
		req->objectname_len = strlen(CADp->ad_filename);
	} else {
		char *tmp = req->query;

		if (req->method != METHOD_GET)
			goto url_error;
		if (TOKEN_EQUAL(req->query, COMMAND_STRING)) {
			tmp += sizeof(COMMAND_STRING)-1;
			return do_command(req, tmp);
		}
		req->objectname = req->query;
		req->objectname_len = req->query_len;
		CADp = parse_GET_cookies(req);
	}

	missed = lookup_urlo(req, LOOKUP_ATOMIC);
	if (req->userspace_module)
		HTTP_BUG();
	urlo = req->urlo;
	if ((!missed && !urlo) || (urlo && urlo->tcapi))
		goto url_error;
	if (req->method == METHOD_GET) {
		if (CADp) {
			ret = custom_ad_rotate(req, CADp);
			if (ret)
				return ret;
		}
	}

	if (missed || !urlo->csumc)
		return http_miss_req(req);

	return ret;

url_error:
	return send_err(req, "CAD: error while parsing CAD request!\n");
}

#define REPLY_HEAD_HEAD \
	"HTTP/1.1 200 OK\r\n" \
	"Content-Type: text/html\r\n" \
	"Connection: Keep-Alive\r\n" \
	"Content-Length: %d\r\n\r\n"

#define REPLY_HEAD_HEAD_COOKIE \
	"HTTP/1.1 200 OK\r\n" \
	"Content-Type: text/html\r\n" \
	"Connection: Keep-Alive\r\n" \
	"Content-Length: %d\r\n" \
	"Set-Cookie: %s\r\n" \
	"\r\n"

#define REPLY_HEAD_TAIL \
	"<html>\n" \
	"<head><title>SPECweb99 Dynamic GET & POST Test</title></head>\n"\
	"<body>\n" \
	"<p>SERVER_SOFTWARE = TUX 1.0\n" \
	"<p>REMOTE_ADDR = %d.%d.%d.%d\n" \
	"<p>SCRIPT_NAME = %s\n" \
	"<p>QUERY_STRING = %s\n" \
	"<pre>\n"

#define REPLY_TAIL \
	"\n</pre>\n" \
	"</body></html>\n"

static int inline send_reply_head (http_req_t *req, size_t body_size)
{
	char buf [1000];
	char *tmp, *head, *tail;
	CAD_t *CADp = (CAD_t *)req->private;
	unsigned int host, head_len, tail_len, total_len;

	host = http_client_addr(req);
#define IP(x) (((unsigned char *)&host)[x])

	tmp = tail = buf;
	tmp += sprintf(tmp, REPLY_HEAD_TAIL, IP(0), IP(1), IP(2), IP(3),
			req->tcapi->vfs_name, req->query);

	tail_len = tmp-buf;

	total_len = tail_len + sizeof(REPLY_TAIL)-1 + body_size;

	head = tmp;
	if (CADp && CADp->reply_cookies_len)
		tmp += sprintf(tmp, REPLY_HEAD_HEAD_COOKIE, total_len, CADp->reply_cookies);
	else
		tmp += sprintf(tmp, REPLY_HEAD_HEAD, total_len);

	head_len = tmp-head;
	http_send_client(req, head, head_len, 0);
	http_send_client(req, tail, tail_len, 0);
	req->http_status = 200;

	return tail_len;
}

static inline int send_reply_tail (http_req_t *req)
{
	int len = sizeof(REPLY_TAIL)-1;

	http_send_client(req, REPLY_TAIL, len, 1);
	return len;
}


/*
 * Send a dynamicly generated buffer. (this is the typical
 * CAD case) Every reply is generated dynamically based on
 * the template and cookie values. The template is scanned
 * for every send.
 */
static int send_reply_CAD (http_req_t *req)
{
	int bytes;
	CAD_t *CADp = (CAD_t *)req->private;

	req->body_len = req->urlo->body_len;

	bytes = send_reply_head(req, req->body_len);
	bytes += scan_send_file(req, CADp);
	bytes += send_reply_tail(req);

	return bytes;
}

/*
 * Return SPECweb99 error message.
 */
static int send_err (http_req_t *req, char *message)
{
	CAD_t *CADp = req->private;
	int len = strlen(message);
	int bytes;

	/*
	 * Return a -1 Ad_id in the reply cookie.
	 */
	CADp->reply_cookies_len = sprintf(CADp->reply_cookies,
			"found_cookie=Ad_id=-1&Ad_weight=0&Expired=1");

	req->body_len = len;
	bytes = send_reply_head(req, len);
	http_send_client(req, message, len, 0);
	bytes += len;
	bytes += send_reply_tail(req);

	req->bytes_sent = bytes;
	return -2;
}

static tcapi_template_t CAD_tcapi = {
	vfs_name: "d",
	version: HTTP_VERSION,
	query: query_CAD,
	send_reply: send_reply_CAD,
};

static int CAD_start (void)
{
	CAD_tcapi.mod = THIS_MODULE;

	return register_httpmodule(&CAD_tcapi);
}

void CAD_stop (void)
{
	unregister_httpmodule(&CAD_tcapi);
}

module_init(CAD_start)
module_exit(CAD_stop)

